#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May  7 14:58:53 2019

@author: mpolikarpov
"""


import numpy as np
import matplotlib.pyplot as plt
import scipy
import cv2


# =============================================================================
# main function for polar interpolation
# =============================================================================

def reproject_image_into_polar(data, origin=None):
    """Reprojects a 3D numpy array ("data") into a polar coordinate system.
    "origin" is a tuple of (x0, y0) and defaults to the center of the image."""
    # horizontal direction on the image will be theta coordinate 
    #(with 0 in the image center), vertical - R coordinate (with 0 in left top corner)
    
    data = cv2.merge([data,data,data])
    ny, nx = data.shape[:2]
    if origin is None:
        origin = (nx//2, ny//2)

    # Determine that the min and max r and theta coords will be...
    x, y = index_coords(data, origin=origin)
    r, theta = cart2polar(x, y)

    # Make a regular (in polar space) grid based on the min and max r & theta
    r_max = int(round(np.sqrt((nx//2)**2 + (ny//2)**2)))
    r_i = np.linspace(r.min(), r.max(), r_max)
    theta_i = np.linspace(theta.min(), theta.max(), 360)
    theta_grid, r_grid = np.meshgrid(theta_i, r_i)

    # Project the r and theta grid back into pixel coordinates
    xi, yi = polar2cart(r_grid, theta_grid)
    xi += origin[0] # We need to shift the origin back to 
    yi += origin[1] # back to the lower-left corner...
    xi, yi = xi.flatten(), yi.flatten()
    coords = np.vstack((xi, yi)) # (map_coordinates requires a 2xn array)

    # Reproject each band individually and the restack
    # (uses less memory than reprojection the 3-dimensional array in one step)
    bands = []
    for band in data.T:
        zi = scipy.ndimage.map_coordinates(band, coords, order=1)
        bands.append(zi.reshape((r_max, 360)))
    output = np.dstack(bands)
    return output#, r_i, theta_i




# =============================================================================
# supplementary functions for polar interpolation
# =============================================================================

def plot_polar_image(data, origin=None):
    """Plots an image reprojected into polar coordinages with the origin
    at "origin" (a tuple of (x0, y0), defaults to the center of the image)"""
    
    polar_grid, r, theta = reproject_image_into_polar(data, origin)
    plt.figure()
    plt.imshow(polar_grid, extent=(theta.min(), theta.max(), r.max(), r.min()))
    plt.axis('auto')
    plt.ylim(plt.ylim()[::-1])
    plt.xlabel('Theta Coordinate (radians)')
    plt.ylabel('R Coordinate (pixels)')
    plt.title('Image in Polar Coordinates')


def index_coords(data, origin=None):
    """Creates x & y coords for the indicies in a numpy array "data".
    "origin" defaults to the center of the image. Specify origin=(0,0)
    to set the origin to the lower left corner of the image."""
    
    ny, nx = data.shape[:2]
    if origin is None:
        origin_x, origin_y = nx // 2, ny // 2
    else:
        origin_x, origin_y = origin
    x, y = np.meshgrid(np.arange(nx), np.arange(ny))
    x -= origin_x
    y -= origin_y
    return x, y


def cart2polar(x, y):
    r = np.sqrt(x**2 + y**2)
    theta = np.arctan2(y, x)
    return r, theta


def polar2cart(r, theta):
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y


def baseline(data1D, polynom = 5):
    '''
    substract the baseline from 2D curve
    '''
    x = np.linspace(0, data1D.shape[0], len(data1D), dtype = 'uint16')
    y = data1D
    
    pfit = np.polyfit(x, y, polynom)                         # returns polynomial coefficients
    yp = np.polyval(pfit, x)                                # fits the curve,
    smoothed = data1D - yp                                  # subtract baseline

    return smoothed


    



# -*- coding: utf-8 -*-
from setuptools import setup


long_description = u"""

Official Github page of the project - https://github.com/maximka48/XIMG-EMBL.

This package is dedicated to image processing, flatifield correction, phase retrieval and X-ray tomography. You may find some examples at GitHub however you may also contact me (Maxim Polikarpov, polikarpov.maxim@mail.ru) in case you would like to use this package and get some guidance. In addition, each function or class have a situational description.

MIT License, Copyright (c) 2019 Maxim Polikarpov
"""


setup(
    name='maximus48',
    version='1.1.7',
    description='Useful tools for image processing & parallel-beam X-ray tomography',
    keywords=['tomography', 'reconstruction', 'imaging'],
    long_description=long_description,
    author='Maxim Polikarpov',
    author_email='polikarpov.maxim@mail.ru',
    url='https://github.com/maximka48/maximus48',
    platforms='OS Independent',
    license='MIT License',
    packages=['maximus48'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Nov 21 16:23:15 2019

@author: mpolikarpov
"""
import os
os.environ['OMP_NUM_THREADS'] ='1'
os.environ['OPENBLAS_NUM_THREADS'] = '1'
os.environ['MKL_NUM_THREADS'] = '1'

from maximus48 import var
import tomopy
import numpy as np
from multiprocessing import Array
from maximus48.var import shift_distance as shift
from maximus48.var import filt_gauss_laplace
from skimage.transform import rescale
from numpy.fft import fft2, fftshift

 
"""
I wanted to make clases F and Processor as light as possible for parallel processing 
so, in a sence, I use them almost as fast dictionaries
"""

# =============================================================================
# additional functions
# =============================================================================

def tonumpyarray(shared_array, shape, dtype):
    '''Create numpy array from shared memory.'''
    nparray = np.frombuffer(shared_array, dtype=dtype).reshape(shape)    
    #assert nparray.base is shared_array
    return nparray
        

def correct_shifts(shifts, median_dev = 5):
    """find any bad numbers which deviate more than 5 pixels from the median
    and correct them to median of the array"""
    
    shifts = tonumpyarray(shifts.shared_array_base, shifts.shape, shifts.dtype)
    
    for i in range(shifts.shape[1]):
        for j in range(shifts.shape[2]):
            shifts[:,i,j] = np.where((abs(shifts[:,i,j] - np.median(shifts[:,i,j])) > median_dev), np.median(shifts[:,i,j]), shifts[:,i,j])
    print('adjusted shifts') 


def fit_polynom(x, y, poly_deg):
    '''fits the y(x) with a cubic function. 
    Returns the y-coordinate of the fit'''

    pfit = np.polyfit(x,y, poly_deg)
    return np.polyval(pfit, x)



# =============================================================================
# # Set of functions for identification of the rotation axis
# =============================================================================
def rotaxis_rough(proj, N_steps = 10):
    """calculate the rotation axis comparing 0 and 180 projection shift
    
    Parameters
    __________
    proj: 3D array
    N_steps: projections per degree
    by default it compares the central part of images (between 1/4 and 3/4 of shape)
    """
    a = proj.shape[1]//4
    b = 3 * proj.shape[1]//4
    c = proj.shape[2]//4
    d = 3 * proj.shape[2]//4
        
                
    cent = []
    N_rot = proj.shape[0] - 180 * N_steps
    
    for i in range(N_rot):
        distances = shift(proj[i, a:b, c:d], np.flip(proj[i + N_steps*180, a:b, c:d] ,1))
        cent.append(proj[i].shape[1]/2 + distances[1]/2)
    
    return cent


def rotaxis_rough_filt(proj, N_steps = 10, sigma = 5, accuracy = 10):
    """calculate the rotation axis comparing 0 and 180 projection shift
    
    Parameters
    __________
    proj: 3D array
    N_steps: projections per degree
    by default it compares the central part of images (between 1/4 and 3/4 of shape)
    """
    a = proj.shape[1]//4
    b = 3 * proj.shape[1]//4
    c = proj.shape[2]//4
    d = 3 * proj.shape[2]//4
        
    cent = []
    N_rot = proj.shape[0] - 180 * N_steps
    
    for i in range(N_rot):
        im1 = filt_gauss_laplace(proj[i, a:b, c:d], sigma)
        im2 = np.flip(filt_gauss_laplace(proj[i + N_steps*180, a:b, c:d], sigma),1)
        distances = shift(im1, im2 , accuracy)
        cent.append(proj[i].shape[1]/2 + distances[1]/2)
    return cent

    

def rotaxis_precise(projections, rotaxis_scan_interval, rot_step = 10, downscale = 0.25):
    """
    This function calculates tomo-reconstructions and tells you 
    which tomo-slice is the sharpest one
    it works like auto-focus in a smartphone
    
    Check this post for the idea (I use FFT)
    https://www.pyimagesearch.com/2015/09/07/blur-detection-with-opencv/
   
    Parameters
    __________
    projections: 3D array
        0 direction is different images
    rotaxis_scan_interval: array
        range of integers for potential rotation axis values
    rot_step: int
        number of radiographic projections per 1 degree (typically 1 or 10)
    downscale: int
        number<1, corresponds to the downscale factor of the image
            for resolution estimation
    elements: 2D array
        0dim: rotaxis coordinate for the most sharpest image
        1dim: standard deviation (contrast) of the tomo-image at this rotaxis
                higher std means better sharpness
    """
            
    # calculate angles
    n = projections.shape[0]
    angle = np.pi*np.arange(n)/(rot_step*180)
     
    # counter for best std
    elements = []
    # 0 direction to store rotaxis
    elements.append([])
    # 1 direction to store contrast values
    elements.append([])    
    
    # body
    for i in rotaxis_scan_interval:
        image = tomopy.recon(projections, angle, center = i, 
                             algorithm = 'gridrec', filter_name = 'shepp')[0]
        
        image = rescale(image, downscale, anti_aliasing=True, multichannel = False)
        image = np.std(np.log(abs(fftshift(fft2(image)))))
        elements[1].append(image)
        elements[0].append(i)
        
    return np.asarray(elements)



def rotaxis_scan(projections, N_slice = 1000, rot_step = 10):
    """
    The function combines rotaxis_rough() and rotaxis_precise()
    It does three iterations to find the best match for the rotation axis
   
    Parameters
    __________
    projections: 3D array
        0 direction is different images. 
        Please note that 1st direction should have 2 or more values. 
        If it has >2 values, only the first slice will be considered 
            for resolution measurements
    rot_step: int
        number of radiographic projections per 1 degree (typically 1 or 10)
    cent: int
        center of rotation
    """
    
    # rough alignment
    cent = rotaxis_rough(projections, rot_step)
    cent = np.median(cent)

    # first iteration
    opa = rotaxis_precise(projections[:,N_slice:N_slice+1], 
                          np.arange(cent - 100, cent + 100, 5), rot_step)
    cent = opa[0, np.argmax(opa[1])]

    # second iteration
    opa = rotaxis_precise(projections[:,N_slice:N_slice+1],
                          np.arange(cent - 5, cent + 5, 1), rot_step)
    cent = opa[0, np.argmax(opa[1])]
    
    # third iteration
    opa = rotaxis_precise(projections[:,N_slice:N_slice+1],
                          np.arange(cent - 2, cent + 2, 0.1), rot_step)
    cent = opa[0, np.argmax(opa[1])]
    
    #final fit
    fit = fit_polynom(opa[0], opa[1], poly_deg = 3)
    cent = opa[0, np.argmax(fit)]
    
    return cent


# =============================================================================
# # Additional custom functions for rotaxis 
# =============================================================================
def axis_raws(image1, image2, Npad = 0, RotROI = None, level = 5, window = 50):
    """Finds an axis of rotation by comparing the 1st and the 180deg image:
    This function is based on rotaxis_rough()
    It is useful when you want to search for rotaxis in different regions of 
        your image independently
        
    Parameters
    __________
    image1: 2D array 
        first 2D image
    image2: 2D array
        second 2D image
    Npad: int 
        type the number here if the x axis of your image was padded
    RotROI: tuple
        ROI where to compare the images. Note that it is better to exclude some regions at the edge of the camera
        Coordinate legend:  RotROI[0] - begin line (if you look at the image, numpy logic)
                            RotROI[1] - end line
                            RotROI[2] - begin column
                            RotROI[3] - end column
    level: int 
        the number of pixels to be taken into account, please type None if you don't want to use it and want to export the whole data.
    window: int
        window is the number of pixels on the image (height) to be taken into account during comaprison
    """
    
    if not RotROI:
        RotROI = (50, image1.shape[0],
                  Npad + image1.shape[1]//8, image1.shape[1] - Npad - image1.shape[1]//8)
    

    all_cent=[]
    for i in range(RotROI[0], RotROI[1], window//2):

        im_1 = image1[i:i+window,RotROI[2]:RotROI[3]]
        im_2 = np.flip(image2[i:i+window,RotROI[2]:RotROI[3]],1)

        distances = shift(im_1, im_2)
        cent = im_1.shape[1]/2 + distances[1]/2 + RotROI[2]

        all_cent.append(cent)
        #print('center for the slice #',i,' is: ', cent)


            
    if level:
        x=[]
        y=[]

        for i in range(len(all_cent)):
            if np.absolute(all_cent[i] - np.median(all_cent)) <level:
                x.append(i * window//2)
                y.append(all_cent[i])
    
        all_cent = np.column_stack((x,y))
    
    else:
        all_cent = np.column_stack((np.linspace(0, len(all_cent), len(all_cent), dtype = 'uint16'), all_cent))
    
    #plt.plot(all_cent[:,0], all_cent[:,1], 'bo')
                
    return all_cent



def interpolate(cent, level = None):
    """interpolates the coordinates for the rotation axis with the line
     basically finds the inclination of the rotation axis through the image.
     The input for this function typically comes from axis_raws() 
    
    Parameters
    __________   
    cent: 2D array 
        comes from axis_raws    
    level:int 
        is the number of pixels to be taken into account if you want 
        to discard all values that have more than 5 degrees difference with the median
    """  

    step = cent[1,0] - cent[0,0]
    
    if not level:
        x = cent[:,0]
        y = cent[:,1]
    
    else:
        x = []
        y = []
        for i in range(len(cent)):
            if np.absolute(cent[i,1] - np.median(cent[:,1])) < level:
                x.append(i*step)
                y.append(cent[i,1])
                
    
    pfit = np.polyfit(x, y, 1)                                                       # returns polynomial coefficients
    #yp = np.polyval(pfit, x)                                                         # fits the curve,
    
    return pfit








# =============================================================================
# Process classesm and functions to define ans store tomo-parameters 
# =============================================================================

#actually should be a part of the Processor class - check tomo_proc.py

def init_Npad(ROI, compression = 8):
    """Calculate the Npad for padding
    can be adjusted with compression parameter
    By default, 8 times smaller than ROI
    """
                    
    if (ROI[2]-ROI[0])>(ROI[3]-ROI[1]):
        Npad = (ROI[2]-ROI[0])//compression       
    else:
        Npad = (ROI[3]-ROI[1])//compression   
        
    return Npad 



def init_names_custom(data_name, distance_indexes):
    """set proper data_names when you manually define indexes for distances"""
    
    data_names = []
    ff_names = []
    
    for i in distance_indexes:
        data_names.append(data_name + '_' + str(i))
        ff_names.append('ff_' + data_name + '_' + str(i))
    
    return data_names, ff_names 
    



# classes themselves

class Processor:
    __slots__ = ['ROI', 'ROI_ff', 'Npad', 'im_shape', 'images', 'flats',
                 'N_files', 'N_start']
        
    def __init__(self, ROI, folder, N_start, N_finish, compNpad = 8):
        """Initialize parameters. 
        Normally should contain ROI, N_distances, etc
        N_distances = tuple
            corresponds to the indexes of the distances
        """
        self.N_start = N_start
        self.ROI = ROI
        self.N_files = (N_finish - N_start) 
        self.im_shape = (ROI[3]-ROI[1], ROI[2]-ROI[0])  
        self.Npad = init_Npad(ROI, compression = compNpad)
        
        
    def init_paths(self, data_name, path, distance_indexes):
        """Generate paths images & flatfields"""
    
        #set data_names
        data_names, ff_names = init_names_custom(data_name = data_name,
                                                 distance_indexes = distance_indexes)
        
        #find images
        imlist = var.im_folder(path)
        
        #set proper paths
        N_distances = len(distance_indexes) 
        images = np.zeros(N_distances, 'object') 
        flats = np.zeros(N_distances, 'object')
                
        for i in np.arange(len(images)):
            #sort image paths
            images[i] = [path+im for im in imlist if (im.startswith(data_names[i])) and not (im.startswith('.'))]
            flats[i] = [path+im for im in imlist if im.startswith(ff_names[i])]
            
        self.images = images
        self.flats = flats

                        

class F:
    __slots__ = ['shape', 'dtype', 'shared_array_base']
    
    def __init__(self, shape, dtype = 'd'):
        """Create shared value array for processing.
        """
        self.shape = shape
        self.dtype = dtype
        
        ncell = int(np.prod(self.shape))
        self.shared_array_base = Array(dtype, ncell,lock=False)       
        pass
     
        
    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 16 15:44:41 2020

@author: mpolikarpov

Computes the Fourier Shell Correlation (FSC) between two given 2d- or
3d-images. In the 2D-setting this is also known as the Fourier Ring
correlation (FRC).

By comparing the FSC between two indepedent reconstructions
of an object from two different data sets to the 1/2-bit-threshold
curve, the achieved resolution can be estimated.

https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2923553/

"""
import numpy as np
from numpy.fft import fftn, ifftshift
from numpy import floor, ceil, real, divide, multiply, sqrt, size




def FSC(im1, im2, beta = 7):
    """
    Calculate Fourier Shell correlation

    Input arguments:
        im1: 2d- or 3d-array containing the first image to be correlated
        im2: size(im1)-array containing the second image to be correlated
        beta: width for the Kaiser Bessel window 
                standard value is beta = 7

    Output arguments:
        frc: length ceil(norm(size(im1))/2) vector containing the computed values
             of the Fourier shell correlation of the images im1, im2, in ascending
             order of frequencies.
        T_hbit: size(fsc)-vector containing the values of the 1/2-bit-threshold
                    curve for the FSC.
        T_bit:  size(fsc)-vector containing the values of the 1-bit-threshold
                    curve for the FSC.
        nu:     size(fsc)-vector containing the Fourier frequencies 
                    corresponding to the values of fsc and t_hbit.
        no_r:   number of bits per ring/shell
    """
    

    N = im1.shape
    ndim = size(N)    

    # Compute Fourier transforms
    # Default case: apply Kaiser-Bessel window prior to Fourier transform
    if beta > 0:
            window = kaiser_bessel(N,beta)
            im1_fft = fftn(im1 * window)
            im2_fft = fftn(im2 * window)
            window = None
    
    # Special case beta == 0: no Kaiser-bessel window applied
    else :  
            im1_fft = fftn(im1)
            im2_fft = fftn(im2)

    # Compute norm of the wave vector for all points in Fourier space
    for jj in range(ndim):
        axis = np.arange(-floor(0.5*N[jj]),  ceil(0.5*N[jj]))**2
        if jj == 0:
            xi = axis
        else:
            xi = xi[..., None] + axis[None, ...] 
    xi = sqrt(ifftshift(xi))

    # Round values to integers for distribution to different Fourier shells
    shells = (np.around(xi)+1).astype('int64')
    
    # Number of Fourier frequencies in each cell
    n_xi = np.bincount(shells.flatten(), 
                       np.ones((size(shells)), dtype = 'int64'))[1:]           # тут получатется нулевой элемент которого нет в матлабе 
    num_xi = size(n_xi)
    
    # Compute correlation on shells
    FRC = divide(
            np.bincount(shells.flatten(), 
                        real(im1_fft.flatten() * np.conj(im2_fft.flatten())),
                        num_xi),
            sqrt(multiply(
                    np.bincount(shells.flatten(),
                                abs(im1_fft.flatten())**2, num_xi),
                    np.bincount(shells.flatten(),
                                abs(im2_fft.flatten())**2, num_xi)
                    )))[1:]
    
    # Restrict to values on shells that are fully contained within the image
    FRC  =  FRC[:int(ceil((min(N)+1)/2))]
    n_xi = n_xi[:int(ceil((min(N)+1)/2))]
    
    # frequencies
    delta_nu = 1/(2*num_xi)
    nu = delta_nu * np.arange(num_xi)
    
    return nu, FRC
    
    
    
     
            
def kaiser_bessel(N, beta):
    """
    make multi-dimensional kaiser-bessel window 
    size(N) corresponds to the number of dimensions in your data
    """
    
    method = np.kaiser
    
    
    # 1D case
    if size(N) == 1:
        window = method(N, beta)
      
    # 2D case
    elif size(N) == 2:
        dim1 = method(N[0],beta)
        dim2 = method(N[1], beta)
        window = dim1[...,None] * dim2[None, ...]
    
    # 3D case
    elif size(N) == 3:
        dim1 = method(N[0],beta)
        dim2 = method(N[1], beta)
        dim3 = method(N[2], beta)
        window = dim1[...,None, None] * dim2[None, ..., None] * dim3[None, None, ...]
    
    else: 
        print('Please check the number of dimensions (N)')
        
    return window












    

